;*********************************************************
; Boot Loader that load another second stage boot loader ;
;*********************************************************

[bits 16]				; we are in 16 bit real mode (all boot loader should has to be in real mode)
[org 0]					; offset is 0 from wherever it is loaded. Boot loeader in loaded at 0x7C00. 
						; We will  set the register later
					
start:  jmp main			; jump to main label to start


;*********************************************************
;               BIOS PARAMETER BLOCK					 ;
;     This is stored at the start of the bootloader		 ;
;*********************************************************

; BPB must begin 3 bytes from the start. We dod a far jmp which is 3 bytes in size
; If we do a short jump, we will nee dto add a 'nop' instruction after to set the offset to 3

bpbOEM:				db 	"My OS   "			; Must be 8 bytes in length. Start offset is 3. That makes the next byte is at the offset 11
bpbBytesPerSector:		dw 	512
bpbSectorsPerCluster:		db	1
bpbReservedSectors:		dw	1				; our boot sector
bpbNumberOfFATs:		db 	2				; 2 FATs table (1 is the copy of the other for backup)
bpbRootEntries:			dw	224				; for floppy disk, there cant be more than 224 directories in the root. There are 2880 sectors in a floppy disk
bpbTotalSectors:		dw	2880
bpbMedia:			db	0xf8				; 11111000 binary: from bit 0: {bit 0} = 0 <=> single side, 
									;                              {bit 1} = 0 => 9 sectors per FAT (1 has 8 secs per Fat)
									;                              {bit 2} = 0 => 0 if it has 80 tracks, 1 if it is 40 tracks. 
									;  			       {bit 3} = 1 => removable; Bits 4 to 7  are unused, and always 1. 
bpbSectorsPerFAT:		dw	9
bpbSectorsPerTrack:		dw 	18
bpbHeadsPerCyclinder:		dw	2
bpbHiddenSectors:		dd	0
bpbTotalSectorsBig:		dd	0

bsDriveNumber:			db	0
bsUnused:			db	0
bsExtBootSignature:		db	0x29
bsSerialNumber:			dd	0xa0a1a2a3
bsVolumeLabel:			db	"MOS FLOPPY "			; must be 11 bytes
bsFileSystem:			db 	"FAT12   "				; must be 8 bytes


;*********************************************************
;               	Entry Point	 		 ;
;*********************************************************

main:
	
	;------------------------------------------------;
	;         Setup segment register value           ;
	;------------------------------------------------;
	
	cli						 ; Important section, disable interrupt
	
	mov ax, 0x07C0					 ; THIS IS 07C0
	mov ds, ax					 ; data segment
	mov es, ax					 ; 3 extra segments
	mov fs, ax
	mov gs, ax
	
	mov ax, 0x0					 ; Stack
	mov ss, ax
	mov sp, 0xFFFF					 ; stack pointer
	
	sti						 ; Restore the interrupt
	
	;------------------------------------------------;
	;         Display the "booting" message          ;
	;------------------------------------------------;
	
	mov si, msgLoading
	Call Print
	
	
	;------------------------------------------------;
	;       Load the ROOT DIRECTORY from disc        ;
	;------------------------------------------------;
	
	; The ReadSectors() needs to know how many sectors to read ==> we need to calculate the size of the ROOT DIRECTORY in sector
	; The ReadSectors() also need the address to start reading from ==> the start of ROOT DIRECTORY
	
							 ; Calculate the size of the root directory in sector and store in cx
							 
	xor cx, cx					 ; Clear cx
	mov ax, 0x0020					 ; no of entries
	
							 ; each entry is 32 byte in size = 0x20	
	mul WORD [bpbRootEntries]			 ; ax = ax * 0x20 
	
	div WORD [bpbBytesPerSector]			 ; Get number of sectors
	
	xchg ax, cx					 ; CX now contains all the number of sectors of the ROOT DIR
	
							
							 ; ---------------------------------------------------------------------
							 ; Calculate the start of the ROOT DIR and store in ax
							 
							 
							 ; [RESERVED SECTOR][FAT TABLES][ROOT DIR] [DATA SECTION]							 
	xor ax, ax
	mov al, BYTE [bpbNumberOfFATs]			 ; number of FAT tables
	mul WORD [bpbSectorsPerFAT]			 ; num of FAT * Sectors Per Fat ==> total sectors for FAT tables
	add ax, WORD [bpbReservedSectors]		 
							
							
							 ;----------------------------------------------------------------------
							 ; once we here, calculate the start of the data section, for later use
	mov WORD [dataSector], ax			 ; for CHS to LBA
	add WORD [dataSector], cx
	
							 ;----------------------------------------------------------------------
							 ; Read the ROOT SECTOR to 7C00:0200
							 ; 7C00:0200 is the sector right after the boot sector
	mov bx, 0x0200
	Call ReadSectors
							 
	;----------------------------------------------------------------------;
	;       Going through the ROOT DIR and looking for KERNEL Image        ;
	;----------------------------------------------------------------------;
		
	mov cx, WORD [bpbRootEntries]			 ; so we know how many entry we need to go through
	mov di, 0x0200					 ; Address of teh 1st entry of the Root DIR in memory
	
     .LOOP:		
	push cx
	push di						 ; di should always point to the first of each entry
	mov cx, 0x000B					 ; first 11 bytes are the file name	
	mov si, KernelImageName
	
	repe cmpsb					 ; test byte to byte while equal and cx > 0
	
	pop di
	pop cx
	
	je LOAD_FAT
	
	add di, 0x0020					 ; add 32 to get to next entry
	
	loop .LOOP
	jmp NO_KERNEL
	
	;-----------------------------------------------;
	;       LOAD FAT        			;
	;-----------------------------------------------;
	
LOAD_FAT:     
							 ; di is the entry: go to byte 26 & 27 to get the first cluster
	mov si, msgNewLine
	Call Print
	
	mov dx, WORD [di + 0x001A]			 ; 0x1A = 26 
	mov WORD [cluster], dx				 ; get the cluster
	
							 ; ----------------------------------------------------------------
							 ; Get the size of the all the FAT tables and stored in cx
	xor ax, ax							 
	mov al, BYTE [bpbNumberOfFATs]							 
	mul WORD [bpbSectorsPerFAT]
	xchg ax, cx
	
							 ; ----------------------------------------------------------------
							 ; Compute the location ax
	mov ax, WORD [bpbReservedSectors]
	
							 ; ----------------------------------------------------------------
							 ; Read FAT and stor in 7C00:0200 (still the second sector)
	mov bx, 0x0200
	Call ReadSectors				 ; We have the first cluster of the kernel image
							 ; to get next cluster, we need the FAT table
			
			
	;------------------------------------------------;
	;      Prepare to read file into 0050:0000	 ;
	;------------------------------------------------;
	mov si, msgNewLine
	Call Print
	
	mov ax, 0x0050					 ; cant mov directly to segment register
	mov es, ax
	mov bx, 0x0000
	
	push bx
	   	
	
LOAD_KERNEL:
	
	mov ax, WORD [cluster]				 ; Cluster to be read	
	Call CHSToLBA
	
	pop bx
	
	xor cx, cx
	mov cl, BYTE [bpbSectorsPerCluster]		 ; num of Sector to read
	
	Call ReadSectors				 ; bx will point to the next buffer
	
	push bx						 ; save the address
	
							 ; Compute the next cluster
							 ; -----------------------------------------------------------------
							 ; Compute the next cluster
	mov ax, WORD [cluster]				 ; entry of this cluster in the FAT table is at index 3/2 of this index
	mov cx, ax
	mov dx, ax

	shr dx, 1					 ; divide dx by 2
	add cx, dx					 ; cx = 3/2 * [cluster]

	mov bx, 0x0200					 ; Location of the FATs in memory
	add bx, cx					 ; index to the FAT table

	mov dx, WORD [bx]				 ; read 2 bytes from FAT 

							 ; cluster is just 12 bits in size

	test ax, 0x0001					 ; test if the current cluster is even or odd
							 ; test like and (not store the value)
							 
	jnz .ODD_CLUSTER
	

   .EVEN_CLUSTER:	

	add dx, 0000111111111111b			 ; remove the top 4 bits (significant)
	jmp .NEXT_CLUSTER
	

   .ODD_CLUSTER:			
   
	shr dx, 0x0004					 ; remove the last 4 bits (least significant)
	

   .NEXT_CLUSTER:
   
	mov WORD [cluster], dx
	cmp dx, 0x0FF0					 ; tset if it is the end of file. last cluster is = 0x0FF0
	jb LOAD_KERNEL
	
   .FINISH_LOAD_KERNEL:
	
	mov si, msgNewLine
	Call Print
	
	push WORD 0x0050
	push WORD 0x0000
	retf
	
NO_KERNEL:	
	
	mov si, msgNoKernelError
	Call Print
	
	mov ah, 0x00					
	int 0x16					 ; wait for a key
	
	int 0x19					 ; reboot
	
;*********************************************************
;               	Other Functions	 		 ;
;*********************************************************	
	
	;------------------------------------------------;
	;           Procedure ReadSectors()	         ;
	;    CX: num of sectors to read			 ;
	;    AX: address of starting sector		 ;
	;    ES:BX: Buffer to store the result		 ;
	;------------------------------------------------;
	
ReadSectors:
   .MAIN:
   	
   	mov di, 0x0005					 ; 5 attempts for an error
   	
   .SECTORLOOP:
   
	push ax						 ; save these registers in case they got modified during the method
	push bx
	push cx

	call LBAToCHS					 ; Convert LBA to CHS
	
							 ; Read Sectors
	mov ah, 0x02					 ; Read Sector function
	mov al, 0x01					 ; Read 1 sector
	mov ch, BYTE [absoluteTrack]			 ; absolute track
	mov cl, BYTE [absoluteSector]
	mov dh, BYTE [absoluteHead]			 ; head
	mov dl, BYTE [bsDriveNumber]			 ; drive
	int 0x13					 ; invoke BIOS for ReadSector 
							 
	jnc .SUCCESS					 ; if CF clear ==> success; CF is set ==> fail

.FAILURE:	

	xor ax, ax					 ; clear ax
	int 0x13					 ; reset disk to first sector when ah = 0
		
	dec di						 ; retry 
	
	pop cx
	pop bx
	pop ax
	
	jnz .SECTORLOOP
	
	int 0x18					 ; finish booting
												 
  .SUCCESS:
	
	mov si, msgProgress				 ; Print a dot, so we know its processing
	Call Print
	
	pop cx						 ; in reverse order
	pop bx
	pop ax	
	
	add bx, WORD [bpbBytesPerSector]		 ; queue for next buffer. Next Sector will be stored right after this sector
	
	inc ax						 ; next sector
	
	loop .MAIN					 ; dec and test cx
	
    	ret
    	
    	
    	
    	
	;---------------------------------------------------------------------------------------;
	;           Procedure LBAToCHS()	         			        	;
	;       AX: LBA address to be converted		 					;
	;       Absolute Sector = (logical sectors MOD sectors per tracks) + 1 			;
	;       Abosulte Head = (logical sectors / sectors per tracks) MOD Num of heads		;
	;											;
	;       Absolute Track = logical sectors / (sectors per tracks * num of Heads)		;
	;		       = logical sectors / sectors per tracks / num of Heads		;
	;---------------------------------------------------------------------------------------;
	
LBAToCHS:
	
							; work out absolute sector
	xor dx, dx					; clean dx
	div WORD [bpbSectorsPerTrack]			; result in ax, remainder in dx
	inc dl
	mov BYTE [absoluteSector], dl
	
							; absolute head
	xor dx, dx
	div WORD [bpbHeadsPerCyclinder]			; remainder in dx
	mov BYTE [absoluteHead], dl
	
	
							; absolute track, 
							; dont need to calc this, the result already in ax
	mov BYTE [absoluteTrack], al			

	ret
	
	;------------------------------------------------;
	;           Procedure CHSToLBA()	         ;
	;     LBA = (cluster - 2) * sec pr clusters	 ;
	;     ax = cluster to be converted
	;------------------------------------------------;
CHSToLBA:
	sub ax, 0x0002
	mul WORD [bpbSectorsPerCluster]
	
	add ax, WORD [dataSector]			 ; base is data sector
	
	ret
	
	;------------------------------------------------;
	;         	Procedure Print()	         ;
	;    SI: address of the string to be displayed   ;
	;          Finish when reaching char 0x00	 ;
	;------------------------------------------------;
	
Print:	
	lodsb						 ; mov 1 byte from si to al
	or al, al					 ; test if it is equal to 0
	jz PrintEnd				         ; if it is, we have finished
	mov ah, 0x0E					 ; print function
	;mov bx, 0x0007					 ; attribute - gray color
	int 10H						 
	jmp Print
	
    PrintEnd:
	ret
	
.data:	
	msgWelcome		db	"Welcome to My OS ", 0x0D, 0x0A, 0x00
	msgLoading 		db	"Loading Boot Image ", 0x0D, 0x0A, 0x00
	msgProgress		db	'.', 0x00
	msgNewLine 		db	0x0D, 0x0A, 0x00
	msgNoKernelError	db 	"ERROR! Press any key to reboot", 0x0D, 0x0A, 0x00
	
	absoluteSector		db	0x00
	absoluteHead		db	0x00
	absoluteTrack		db	0x00
	
	dataSector		dw	0x0000
	
	cluster 		dw	0x0000
	
	KernelImageName		db	"KRNLDR  SYS"    	; must be 11 bytes



TIMES 510 - ($ - $$) db 0
dw 0xAA55


















